from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from .constants import INTERVAL_OF, OFFSET_OF

if TYPE_CHECKING:
    from .chroma import Chroma


class Interval:
    def __init__(
        self, value: Union[int, str, "Interval", "Chroma"], octave: Optional[int] = 0
    ):
        # Avoid circular imports
        from .chroma import Chroma

        match value:
            case int():
                self.distance = value
            case str():
                self.distance = OFFSET_OF[value] + octave * 12
            case Interval():
                self.distance = value.distance
            case Chroma():
                self.distance = value.offset
            case _:
                raise TypeError("expected value of type int|str|Interval|Chroma")

    @property
    def offset(self) -> int:
        """Get the interval offset within an octave."""
        return self.distance % 12

    @offset.setter
    def offset(self, value: int) -> None:
        """Set the interval offset while preserving the octave."""
        self.distance = value + self.octave * 12

    @property
    def octave(self) -> int:
        """Get the octave component of the interval."""
        return self.distance // 12

    @octave.setter
    def octave(self, value: int) -> None:
        """Set the octave while preserving the offset."""
        self.distance = value * 12 + self.offset

    @property
    def interval(self) -> str:
        """Get the interval name assuming the same octave."""
        return INTERVAL_OF[self.offset]

    @interval.setter
    def interval(self, value: str) -> None:
        """Set the interval by name assuming the same octave."""
        self.offset = OFFSET_OF[value]

    def __str__(self) -> str:
        return f"{self.interval}({self.octave})" if self.octave else self.interval

    def __repr__(self) -> str:
        return f"Interval('{self.interval}', octave={self.octave})"

    def __invert__(self) -> int:
        """Return the distance value of the interval."""
        return self.distance

    def __hash__(self) -> int:
        return hash(str(self))

    def __eq__(self, other: Union[int, str, Interval]) -> bool:
        return self.distance == Interval(other).distance

    def __gt__(self, other: Union[int, str, Interval]) -> bool:
        return self.distance > Interval(other).distance

    def __ge__(self, other: Union[int, str, Interval]) -> bool:
        return self > other or self == other

    def __lt__(self, other: Union[int, str, Interval]) -> bool:
        return not (self >= other)

    def __le__(self, other: Union[int, str, Interval]) -> bool:
        return not (self > other)

    def __add__(self, other: Union[int, str, Interval]) -> Interval:
        return Interval(self.distance + Interval(other).distance)

    def __sub__(self, other: Union[int, str, Interval]) -> Interval:
        return Interval(self.distance - Interval(other).distance)


# shorthand
I = Interval



from .chroma import C, Chroma
from .constants import ABC_OF, CHROMA_OF, INTERVAL_OF, OFFSET_OF
from .interval import I, Interval
from .pitch import P, Pitch
from .rest import Rest
from .voicing import V, Voicing

__all__ = [
    "Rest",
    "CHROMA_OF",
    "ABC_OF",
    "INTERVAL_OF",
    "OFFSET_OF",
    "Interval",
    "I",
    "Chroma",
    "C",
    "Pitch",
    "P",
    "Voicing",
    "V",
]



class Rest(int):
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls, 0)
        return cls._instance

    def __init__(self):
        pass

    def __repr__(self):
        return 'Rest()'

    def __str__(self):
        return ''

    def __add__(self, other):
        return self

    def __radd__(self, other):
        return self

    def __sub__(self, other):
        return self

    def __rsub__(self, other):
        return self

    def __mul__(self, other):
        return self

    def __rmul__(self, other):
        return self

    def __truediv__(self, other):
        return self

    def __rtruediv__(self, other):
        return self

    def __floordiv__(self, other):
        return self

    def __rfloordiv__(self, other):
        return self

    def __mod__(self, other):
        return self

    def __rmod__(self, other):
        return self

    def __pow__(self, other, modulo=None):
        return self

    def __rpow__(self, other):
        return self

    def __and__(self, other):
        return self

    def __or__(self, other):
        return self

    def __xor__(self, other):
        return self

    def __lshift__(self, other):
        return self

    def __rshift__(self, other):
        return self

    def __eq__(self, other):
        return isinstance(other, Rest)

    def __ne__(self, other):
        return not isinstance(other, Rest)

    def __lt__(self, other):
        return False

    def __le__(self, other):
        return False

    def __gt__(self, other):
        return False

    def __ge__(self, other):
        return False

    def __neg__(self):
        return self

    def __pos__(self):
        return self

    def __abs__(self):
        return self

    def __invert__(self):
        return self

    def __hash__(self):
        return 9999



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Union

from .constants import CHROMA_OF, OFFSET_OF
from .interval import Interval

if TYPE_CHECKING:
    from .pitch import Pitch


class Chroma:
    """Represents a pitch class (chroma) without a specific octave."""

    def __init__(self, value: Union[int, str, "Chroma", "Pitch"]):
        # Import here to avoid circular imports
        from .pitch import Pitch

        match value:
            case int():
                self.offset = value % 12
            case str():
                self.offset = OFFSET_OF[value]
            case Chroma():
                self.offset = value.offset
            case Pitch():
                self.offset = value.offset
            case _:
                raise TypeError("expected value of type int|str|Chroma|Pitch")

    def __str__(self) -> str:
        return CHROMA_OF[self.offset]

    def __repr__(self) -> str:
        return f"Chroma('{self}')"

    def __invert__(self) -> int:
        """Return the offset value of the chroma."""
        return self.offset

    def __eq__(self, other: Any) -> bool:
        from .pitch import Pitch

        match other:
            case int():
                return self.offset == other % 12
            case str():
                return self.offset == OFFSET_OF[other]
            case Chroma():
                return self.offset == other.offset
            case Pitch():
                return self.offset == other.offset
            case _:
                return False

    def __neg__(self) -> int:
        return -self.offset

    def __mul__(self, value: Union[int, str, Interval]) -> Chroma:
        """Transpose chroma upwards."""
        return Chroma((Interval(value) + self.offset).offset)

    def __add__(self, value: Union[int, str, Interval]) -> Chroma:
        """Transpose chroma upwards (alias for __mul__)."""
        return Chroma((Interval(value) + self.offset).offset)

    def __rshift__(self, value: Union[int, str, Interval]) -> Chroma:
        """Transpose chroma upwards (alias for __mul__)."""
        return Chroma((Interval(value) + self.offset).offset)

    def __truediv__(self, value: Union[int, str, Interval]) -> Chroma:
        """Transpose chroma downwards."""
        return Chroma(self.offset - Interval(value).distance)

    def __sub__(self, value: Union[int, str, Interval, Chroma]) -> Union[int, Chroma]:
        """Transpose downwards or compute difference between chromas."""
        if isinstance(value, Chroma):
            return (self.offset - value.offset) % 12
        return Chroma(self.offset - Interval(value).distance)

    def __lshift__(self, value: Union[int, str, Interval]) -> Chroma:
        """Transpose chroma downwards (alias for __truediv__)."""
        return Chroma(self.offset - Interval(value).distance)

    def __rsub__(self, pitches: List[Pitch]) -> List[Pitch]:
        """Filter out pitches that match this chroma."""
        from .pitch import Pitch

        return [Pitch(_) for _ in pitches if _ != self]


# shorthand
C = Chroma



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, Union

from .chroma import Chroma
from .constants import ABC_OF, CHROMA_OF, OFFSET_OF

if TYPE_CHECKING:
    from .interval import Interval


class Pitch:
    def __init__(self, value: Union[int, str, Chroma, Pitch], octave: int = 4):
        match value:
            case int():
                self.value = value
            case str():
                if any(_ in value for _ in '_=^'):
                    self.abc = value
                    return
                chroma = ''.join(_ for _ in value if _.isalpha() or _ == '#')
                if len(chroma) < len(value):
                    octave = int(value[len(chroma) :])
                self.value = OFFSET_OF[chroma] + (octave + 1) * 12
            case Chroma():
                self.value = value.offset + (octave + 1) * 12
            case Pitch():
                self.value = value.value
            case _:
                raise TypeError("expected value of type int|str|Chroma|Pitch")

    def __str__(self) -> str:
        return f"{self.name}{self.octave}"

    def __repr__(self) -> str:
        return f"Pitch('{self.name}', {self.octave})"

    def __invert__(self) -> int:
        """Return the MIDI value of the pitch."""
        return self.value

    def __hash__(self) -> int:
        return hash(str(self))

    def __lt__(self, other: Any) -> bool:
        match other:
            case int():
                return self.value < other
            case Pitch():
                return self.value < other.value
            case _:
                return NotImplemented

    def __le__(self, other: Any) -> bool:
        match other:
            case int():
                return self.value <= other
            case Pitch():
                return self.value <= other.value
            case _:
                return NotImplemented

    def __gt__(self, other: Any) -> bool:
        match other:
            case int():
                return self.value > other
            case Pitch():
                return self.value > other.value
            case _:
                return NotImplemented

    def __ge__(self, other: Any) -> bool:
        match other:
            case int():
                return self.value >= other
            case Pitch():
                return self.value >= other.value
            case _:
                return NotImplemented

    def __eq__(self, other: Any) -> bool:
        from .interval import Interval

        match other:
            case int():
                return self.value == other
            case str():
                return self.name == other
            case Interval():
                return self.offset == other.distance
            case Chroma():
                return self.offset == other.offset
            case Pitch():
                return self.value == other.value
            case _:
                return False

    def __mul__(self, n: int) -> list[Pitch]:
        """Return a list of copies."""
        return [Pitch(self) for i in range(n)]

    def __rshift__(
        self, interval: Union[int, str, "Interval", Chroma, "Pitch"]
    ) -> Pitch:
        """Transpose pitch upwards."""
        from .interval import Interval

        match interval:
            case Pitch():
                return Pitch(self.value + interval.value)
            case _:
                return Pitch(self.value + Interval(interval).distance)

    #    def __truediv__(self, interval: Union[int, str, 'Interval', Chroma, 'Pitch']) -> Pitch:
    #        """Transpose pitch downwards."""
    #        from .interval import Interval
    #        match interval:
    #            case Pitch():
    #                return Pitch(self.value - interval.value)
    #            case _:
    #                return Pitch(self.value - Interval(interval).distance)

    def __lshift__(
        self, interval: Union[int, str, "Interval", Chroma, "Pitch"]
    ) -> Pitch:
        """Transpose pitch downwards."""
        from .interval import Interval

        match interval:
            case Pitch():
                return Pitch(self.value - interval.value)
            case _:
                return Pitch(self.value - Interval(interval).distance)

    def __add__(self, other: Union[int, "Pitch", List["Pitch"]]) -> List["Pitch"]:
        """Concat pitch with other pitch(es)."""
        match other:
            case int() | Pitch():
                return sorted([Pitch(self), Pitch(other)])
            case list():
                return sorted([Pitch(self)] + [Pitch(_) for _ in other])
            case _:
                raise TypeError("expected value of type int|Pitch|list[Pitch]")

    def __radd__(self, other: Union[int, "Pitch", List["Pitch"]]) -> List["Pitch"]:
        """Reverse add operation."""
        return self.__add__(other)

    def __sub__(self, other: Union[int, "Pitch"]) -> int:
        """Compute interval between pitches."""
        return self.value - Pitch(other).value

    def __rsub__(self, pitches: List["Pitch"]) -> List["Pitch"]:
        """Filter out this pitch from a list."""
        return [Pitch(_) for _ in pitches if _ != self]

    @property
    def offset(self) -> int:
        """Get the offset within an octave."""
        return self.value % 12

    @offset.setter
    def offset(self, value: int) -> None:
        """Set the offset while preserving the octave."""
        self.value = (self.octave + 1) * 12 + value

    @property
    def octave(self) -> int:
        """Get the octave of the pitch."""
        return self.value // 12 - 1

    @octave.setter
    def octave(self, value: int) -> None:
        """Set the octave while preserving the offset."""
        self.value = self.offset + (value + 1) * 12

    @property
    def name(self) -> str:
        """Get the pitch name."""
        return CHROMA_OF[self.offset]

    @name.setter
    def name(self, value: str) -> None:
        """Set the pitch by name."""
        self.offset = OFFSET_OF[value]

    @property
    def chroma(self) -> Chroma:
        """Get the chroma of this pitch."""
        return Chroma(self)

    @chroma.setter
    def chroma(self, value: Chroma) -> None:
        """Set the chroma while preserving the octave."""
        self.offset = value.offset

    @property
    def abc(self) -> str:
        """Get the ABC notation of this pitch."""
        abc = ABC_OF[self.offset]
        if abc == "z":
            return abc
        if (va := self.octave - 4) <= 0:
            return abc + "," * -va
        return abc.lower() + "'" * (va - 1)

    @abc.setter
    def abc(self, abc: str) -> None:
        """Set the ABC notation of this pitch."""
        if abc == 'z':
            self.value = Rest()
            return
        name, suffix = (abc[0], abc[1:]) if abc[0].isalpha() else (abc[:2], abc[2:])
        if va := int(name[-1].islower()):
            name = name.upper()
        va += suffix.count("'") - suffix.count(",")
        self.value = 0
        self.offset = OFFSET_OF[name]
        self.octave = 4 + va

    @classmethod
    def from_abc(cls, abc: str):
        """Create a new pitch from ABC notation"""
        pitch = cls(0)
        pitch.abc = abc
        return pitch


# shorthand
P = Pitch



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Union

from .chroma import Chroma
from .constants import OFFSET_OF
from .pitch import Pitch

if TYPE_CHECKING:
    from .interval import Interval


class Voicing:
    """Represents a musical voicing (collection of pitches possibly with a root)."""

    def __init__(
        self,
        pitches: Union[list[Union[int, str, Chroma, Pitch]], str] = [],
        root: Union[int, str, Chroma, Pitch, Voicing, None] = None,
    ):
        if isinstance(pitches, str):
            pitches = pitches.split(" ")
        self.pitches = [Pitch(_) for _ in pitches]
        if root is None and isinstance(pitches, Voicing):
            self.root = pitches.root
        else:
            self.root = root

    @property
    def root(self) -> Chroma:
        return self._root

    @root.setter
    def root(self, value: Union[int, str, Chroma, Pitch, None]) -> None:
        self._root = None if value is None else Chroma(value)

    @property
    def int_list(self):
        return [_.value for _ in self]

    @property
    def csv(self) -> str:
        return ",".join(str(_.value) for _ in self)

    @csv.setter
    def csv(self, csv: str, sep=",") -> None:
        self.pitches = [Pitch(int(_)) for _ in csv.split(sep)]

    @classmethod
    def from_csv(cls, csv: str):
        v = cls()
        v.csv = csv
        return v

    def __str__(self) -> str:
        return f'{self.root}{self.quality()}[{" ".join([str(_) for _ in self])}]'

    def __repr__(self) -> str:
        if self.root is None:
            return f"Voicing('{' '.join(str(_) for _ in self)}')"
        return f"Voicing('{' '.join(str(_) for _ in self)}', '{self.root}')"

    def __hash__(self) -> int:
        return hash(str(self))

    def __len__(self) -> int:
        """Get the number of pitches in the voicing."""
        return len(self.pitches)

    def __iter__(self):
        """Iterate through the pitches in the voicing."""
        return iter(self.pitches)

    def __eq__(self, other: Any) -> bool:
        """Compare if two voicings are equivalent (taking into account transposition)."""
        if not isinstance(other, Voicing) or len(self) != len(other):
            return False
        if len(self) == 0:
            return True
        # Transpose to the same root and compare pitches
        return (self << (self[0] - other[0])).pitches == other.pitches

    def __gt__(self, other: Voicing) -> bool:
        """Check if this voicing is a superset of the other."""
        for pitch in other:
            if pitch not in self:
                return False
        return len(self) > len(other)

    def __ge__(self, other: Voicing) -> bool:
        """Check if this voicing is a superset of or equal to the other."""
        return self > other or self == other

    def __mod__(self, value: Union[int, str]) -> list[Pitch]:
        """Find pitches that form the given interval with higher pitches."""
        if isinstance(value, str):
            value = OFFSET_OF[value]

        lowers = []
        for i in range(len(self) - 1):
            for j in range(i + 1, len(self)):
                if self[j] - self[i] == value:
                    lowers.append(Pitch(self[i]))
        return lowers

    def __add__(self, value: Union[int, str, Pitch, list[Pitch], Voicing]) -> Voicing:
        """Add pitch(es) to the voicing."""
        ret = Voicing(self)
        match value:
            case int() | str() | Pitch():
                ret.pitches.append(Pitch(value))
            case list() | Voicing():
                for pitch in value:
                    ret.pitches.append(Pitch(pitch))
            case _:
                raise TypeError(
                    "expected value of type int|str|Pitch|list[Pitch]|Voicing"
                )
        ret.pitches.sort()
        return ret

    def __sub__(self, value: Union[str, Pitch, list[Pitch], Voicing]) -> Voicing:
        """Remove pitch(es) from the voicing."""
        ret = Voicing(self)
        match value:
            case str():
                p = Pitch(value)
                if p in self:
                    ret.pitches.remove(p)
            case Pitch():
                if value in self:
                    ret.pitches.remove(value)
            case list():
                for v in value:
                    p = Pitch(v)
                    if p in self:
                        ret.pitches.remove(p)
            case Voicing():
                for pitch in value:
                    if pitch in self:
                        ret.pitches.remove(pitch)
            case _:
                raise TypeError("expected value of type str|Pitch|list[Pitch]|Voicing")
        return ret

    def __mul__(self, n: int) -> list[Voicing]:
        """Return a list of copies."""
        return [Voicing(self) for i in range(n)]

    def __rshift__(self, value: Union[int, str, "Interval"]) -> Voicing:
        """Transpose the voicing upwards."""
        return Voicing(
            [_ >> value for _ in self],
            None if self.root is None else self.root >> value,
        )

#    def __truediv__(self, value: Union[int, str, 'Interval', Chroma]) -> Voicing:
#        """Transpose the voicing downwards."""
#        from .interval import Interval
#        ret = Voicing(self)
#        ret.root -= value
#        for i in range(len(ret)):
#            ret[i] /= value
#        return ret

    def __lshift__(self, value: Union[int, str, "Interval", Chroma]) -> Voicing:
        """Transpose the voicing downwards."""
        return self.__rshift__(-value)

    def __floordiv__(self, target: Union[int, Pitch]) -> Voicing:
        """Transpose the voicing to a specific target pitch/value."""
        if len(self) == 0:
            return Voicing(self)
        root = self[0] << (self[0].chroma - self.root)
        return self << (root - target)

    def matrix(self) -> list[list[int]]:
        """Compute the interval matrix between all pitches in the voicing.

        Returns:
            A 2D array where each cell represents the interval between two pitches
        """
        if len(self) <= 1:
            return []
        interval_matrix = []
        for i in range(len(self) - 1):
            interval_matrix.append([])
            for j in range(len(self)):
                interval = self[j] - self[i] if j > i else 0
                interval_matrix[i].append(interval)
        return interval_matrix

    def index(self, key: Union[int, str, Pitch, Chroma]) -> int:
        if isinstance(key, str):
            if any(_.isdigit() for _ in key):
                # Look for a pitch name with octave like "C4"
                return (~self).index(key) if key in ~self else -1
            # Look for a chroma name
            key = Chroma(key)

        try:
            return self.pitches.index(key)
        except ValueError:
            return -1

    def __getitem__(self, key: Union[int, str, Pitch, Chroma]) -> Optional[Pitch]:
        """Get a pitch by index, name, or object.

        Args:
            key: The index or pitch to retrieve

        Returns:
            The pitch at the given index or None if not found
        """
        if isinstance(key, int):
            return self.pitches[key]
        idx = self.index(key)
        return self.pitches[idx] if idx >= 0 else None

    def __delitem__(self, key: Union[int, str, Pitch, Chroma]) -> None:
        """Remove a pitch from the voicing.

        Args:
            key: The index or pitch to remove
        """
        if isinstance(key, int):
            del self.pitches[key]
        else:
            idx = self.index(key)
            if idx >= 0:
                del self.pitches[idx]

    def __setitem__(
        self, key: Union[int, str, Pitch, Chroma], value: Union[int, Pitch]
    ) -> None:
        """Set a pitch at the given index or replace a pitch.

        Args:
            key: The index or pitch to replace
            value: The new pitch
        """
        if isinstance(key, int):
            self.pitches[key] = Pitch(value)
        else:
            idx = self.index(key)
            if idx >= 0:
                self.pitches[idx] = Pitch(value)

    @property
    def abc(self) -> list[str]:
        return [_.abc for _ in self]

#    def voice(self, pitch: Pitch):
#        """Closest voice to the given pitch"""
#        if len(self) == 0:
#            return None
#        distance_from_pitch = lambda i: abs(self.pitches[i]-pitch)
#        return min(range(len(self)), key=distance_from_pitch)

    @property
    def drop2(self):
        assert len(self) == 4
        return Voicing((self[2]<<12, self[0], self[1], self[3]), self.root)

    @property
    def drop3(self):
        assert len(self) == 4
        return Voicing((self[1]<<12, self[0], self[2], self[3]), self.root)

    @property
    def drop24(self):
        assert len(self) == 4
        return Voicing((self[0]<<12, self[2]<<12, self[1], self[3]), self.root)

    def __invert__(self) -> list[str]:
        """Analyze the chord tones relative to the root."""
        if len(self) == 0:
            return []
        offsets = [_.offset for _ in (self << self.root)]
        has = lambda *values: any(_ in offsets for _ in values)
        tones = [
            {
                0: "1",
                1: "b9",
                2: "9" if has(3, 4) else "sus2",
                3: "#9" if has(4) else "min3",
                4: "maj3",
                5: "11" if has(3, 4) else "sus4",
                6: "#11" if has(4, 7) else "b5",
                7: "5",
                8: "b13" if has(3, 7, 10) else "#5",
                9: "13" if has(10, 11) else "6",
                10: "min7" if has(3) else "dom7",
                11: "maj7",
            }[_]
            for _ in offsets
        ]
        if "b5" in tones and "6" in tones:
            tones = ["dim7" if _ == "6" else _ for _ in tones]
        return tones

    def quality(self) -> str:
        return ""
#        if self.root is None or len(self) < 2:
#            return ''
#        tones = ~self
#        base = ''
#        tensions = ''
#        return base + tensions


# shorthand
V = Voicing
